/*
Copyright © 2021-2023 Macaroni OS Linux
See AUTHORS and LICENSE for the license details and contributors.
*/
package utils

import (
	"io"
	"io/fs"
	"os"
)

func KeyInList(key string, arr *[]string) bool {
	for _, k := range *arr {
		if k == key {
			return true
		}
	}

	return false
}

// Exists reports whether the named file or directory exists.
func Exists(name string) bool {
	if _, err := os.Stat(name); err != nil {
		if os.IsNotExist(err) {
			// Check if the file is a broken link.
			link, err := os.Readlink(name)
			// If the file is not a link an error
			// readlink <path>: invalid argument
			// is generated.
			if link != "" && err == nil {
				return true
			}
			return false
		} else {
			// Check if the file is a broken link.
			link, err := os.Readlink(name)
			// If the file is not a link an error
			// readlink <path>: invalid argument
			// is generated.
			if link != "" && err == nil {
				return true
			}
			return false
		}
	}
	return true
}

func IsDir(path string) (bool, error) {
	fileInfo, err := os.Stat(path)
	if err != nil {
		return false, err
	}
	return fileInfo.IsDir(), err
}

func IsLink(path string) (bool, error) {
	fileInfo, err := os.Lstat(path)
	if err != nil {
		return false, err
	}
	if fileInfo.Mode()&fs.ModeSymlink != 0 {
		return true, nil
	}
	return false, nil
}

// Copy copies the contents of the file at srcpath to a regular file at dstpath.
// If dstpath already exists and is not a directory, the function truncates it.
// The function does not copy file modes or file attributes.
func CopyFile(srcpath, dstpath string) (err error) {
	r, err := os.Open(srcpath)
	if err != nil {
		return err
	}
	defer r.Close() // ok to ignore error: file was opened read-only.

	w, err := os.Create(dstpath)
	if err != nil {
		return err
	}

	defer func() {
		c := w.Close()
		// Report the error from Close, if any.
		// But do so only if there isn't already
		// an outgoing error.
		if c != nil && err == nil {
			err = c
		}
	}()

	_, err = io.Copy(w, r)
	return err
}
